package dareka;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.net.BindException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.MalformedURLException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.channels.ServerSocketChannel;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.regex.Pattern;

import dareka.common.CloseUtil;
import dareka.common.Config;
import dareka.common.Logger;
import dareka.extensions.Extension;
import dareka.extensions.RequestFilter;
import dareka.extensions.Rewriter;
import dareka.processor.Processor;
import dareka.processor.impl.CacheDirProcessor;
import dareka.processor.impl.ConnectProcessor;
import dareka.processor.impl.DefaultRequestFilter;
import dareka.processor.impl.GetPostProcessor;
import dareka.processor.impl.LocalDirProcessor;
import dareka.processor.impl.NGWordProcessor;
import dareka.processor.impl.NicoCachingProcessor;
import dareka.processor.impl.RewriterProcessor;
import dareka.processor.impl.ThumbProcessor;

public class Server {
    private static final int MAX_WAITING_TIME = 10;

    private Config config;
    private ServerSocket serverSocket;
    private ExecutorService executor = Executors.newCachedThreadPool();
    private boolean stopped = false;

    // they are able to be shared among threads.
    private Processor connectProcessor = new ConnectProcessor();
    private Processor getPostProcessor = new GetPostProcessor();
    private RewriterProcessor rewriterProcessor = null;

    public Server(Config config) throws IOException {
        if (config == null) {
            throw new IllegalArgumentException("config must not be null");
        }

        this.config = config;

        // use channel to make it available Socket#getChannel() for non blocking
        // I/O.
        ServerSocketChannel serverCh = ServerSocketChannel.open();
        serverSocket = serverCh.socket();
        
        //[nl] gProcessor̃[h
        rewriterProcessor = new RewriterProcessor(config);
        initClassLoader();
    }

    /**
     * Start the server. The thread which call this method is blocked until
     * stop() is called or some errors occurred.
     */
    public void start() {
        if (stopped) {
            return;
        }

        try { // ensure cleanup
            bindServerSocket();
            acceptServerSocket();
        } finally {
            Logger.info("finalizing");
            cleanupServerSocket();
            cleanupExecutor();
        }
    }

    /**
     * Stop the server. Please call this method from another thread which called
     * start().
     */
    public synchronized void stop() {
        stopped = true;

        if (!serverSocket.isClosed()) {
            CloseUtil.close(serverSocket);
        }
        executor.shutdown();
    }

    private void bindServerSocket() {
		int port = Integer.getInteger("listenPort").intValue();
		try {
			// [nl] IP
			if(config.allowFrom.equals("all")
					|| config.allowFrom.startsWith("lan")) {
				serverSocket.bind(new InetSocketAddress(port));
			} else {
				serverSocket.bind(new InetSocketAddress(
						InetAddress.getByName(null), port));
			}
		} catch (BindException e) {
			Logger.info("|[g %d ɎgĂ邽ߋNł܂B", port);
			stop();
        } catch (IOException e) {
            Logger.error(e);
            stop();
        }
    }

    private void acceptServerSocket() {
        try {
            boolean timeoutSupportedOrUnknown = true;
            int timeout = Config.getInteger("readTimeout", 600000);
            
            while (!stopped) {
                Socket client = serverSocket.accept();

                try { // ensure client.close() even in errors.
                	if (timeoutSupportedOrUnknown) {
                        client.setSoTimeout(timeout);
                        if (client.getSoTimeout() != timeout) {
                            Logger.warning("read timeout is not suported");
                            timeoutSupportedOrUnknown = false;
                        }
                    }
                	
	                synchronized (this) { // avoid to conflict with stop()
	                    if (stopped) {
	                        break;
	                    }
	                    
						// [nl] lan̔
						if (config.allowFrom.startsWith("lan"))
						{
							String lanClass = config.allowFrom.substring(3);
							byte[] ip = client.getInetAddress().getAddress();
							if (client.getInetAddress().isLoopbackAddress()
									|| (lanClass.equals("A") && ip[0] == 10)
									|| (lanClass.equals("B") && ip[0] == -84/*172*/ && 16<= ip[1] && ip[1] < 32)
									|| (lanClass.equals("C") && ip[0] == -64/*192*/ && ip[1] == -88/*168*/))
							{
								;//ok
							}
							else
							{
								Logger.info("Forbidden address: " + client.getInetAddress().getHostAddress());
								client.close();
								continue;
							}
						}
	
                        ConnectionManager worker;
                        worker = new ConnectionManager(config, client);
	                    // TODO R[fBOXœo^ł悤ɂB
	
	                    registerProcessor(new NicoCachingProcessor(executor), worker);
	                    
	                    //[nl] ftHgtB^̓o^
	                    worker.addRequestFilter(new DefaultRequestFilter());
	                    
	                    // [nl] gn̓o^
	                    registerExtProcessors(worker);
	
	                    // [nl] Processor̓o^
	                    registerProcessor(new LocalDirProcessor(config), worker);
	                    registerProcessor(new CacheDirProcessor(), worker);
	                    registerProcessor(rewriterProcessor, worker);
	                    registerProcessor(new NGWordProcessor(config), worker);
	                    
	                    if (config.cacheThumbnail)
	                    	registerProcessor(new ThumbProcessor(), worker);
	                    
	                    registerProcessor(getPostProcessor, worker);
	                    registerProcessor(connectProcessor, worker);

	                    executor.execute(worker);
                    	// for debug
	                    //new Thread(worker).start();
	                }
	            } catch (Exception e) {
                    Logger.error(e);
                    CloseUtil.close(client);
                }
            }
        } catch (IOException e) {
            // stop() is called.
            // including AsynchronousCloseException (in NIO)
            Logger.debug(e);
        }
    }
    
    // [nl] gProcessor̃eXg
    private URLClassLoader extLoader;
    private List<Extension> procExtensions = new ArrayList<Extension>();
    private List<Extension> filtExtensions = new ArrayList<Extension>();
    private final Extension.Type supportedProcType = Extension.Type.Processor1;
    //private final Extension.Type supportedReWrType = Extension.Type.Rewriter1;
    private final Extension.Type supportedFiltType = Extension.Type.RequestFilter1;
    private void initClassLoader()
    {
    	URL[] urls = new URL[1];
    	try {
			urls[0] = new File("").toURI().toURL();
		} catch (MalformedURLException e) { }
        extLoader = new URLClassLoader(urls);
        
        File extDir = new File("extensions");
        File[] files = extDir.listFiles(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                if (name.matches(".+\\.class")) {
                    return true;
                } else {
                    return false;
                }
            }
        });

        if (files == null)
        	return ;
        
        Arrays.sort(files);
        for (File file: files)
        {
        	try {
        		String className = file.getName().replace(".class", "");
				Object extObj = extLoader.loadClass(
								"extensions." + className).newInstance();
				Extension extInterface = null;
				if (extObj instanceof Extension) {
					extInterface = (Extension)extObj;
				} else {
					continue;
				}
				
				Object obj = extInterface.queryInterface(Extension.Type.Processor1);
				if (obj instanceof Processor)
				{
					Logger.info("Processor Extension: " + extInterface.getVersionString());
					System.setProperty("extension." + className, "true");
					procExtensions.add(extInterface);
				}
				obj = extInterface.queryInterface(Extension.Type.Rewriter1);
				if (obj instanceof Rewriter)
				{
					Logger.info("Rewriter Extension: " + extInterface.getVersionString());
					System.setProperty("extension." + className, "true");
					rewriterProcessor.addRewriter((Rewriter)obj);
				}
				obj = extInterface.queryInterface(Extension.Type.RequestFilter1);
				if (obj instanceof RequestFilter)
				{
					Logger.info("RequestFilter Extension: " + extInterface.getVersionString());
					System.setProperty("extension." + className, "true");
					filtExtensions.add(extInterface);
				}
			} catch (ClassNotFoundException e) {
				Logger.warning("Ext Error(NoSuchClass): " + e.getMessage());
			} catch (InstantiationException e) {
			} catch (IllegalAccessException e) {
			} catch (AbstractMethodError e) {
				Logger.warning("Ext Error(NoSuchMethod): " + e.getMessage());
			}
        }
    }
    
    private void registerExtProcessors(ConnectionManager worker)
    {
    	// Processor
    	for (Extension entry : procExtensions) {
    		registerProcessor(
    				(Processor)entry.queryInterface(supportedProcType),
    				worker);
    	}
    	
    	// RequestFilter
    	for (Extension entry : filtExtensions) {
    		worker.addRequestFilter((RequestFilter)entry.queryInterface(supportedFiltType));
    	}
    }

    private void registerProcessor(Processor processor, ConnectionManager worker) {
        Pattern p = processor.getSupportedURLAsPattern();
        if (p == null) {
            String url = processor.getSupportedURLAsString();
            if (url != null) {
                p = Pattern.compile(url, Pattern.LITERAL);
            }
        }

        String[] methods = processor.getSupportedMethods();
        if (methods == null) {
            return;
        }

        for (String method : methods) {
            worker.addProcessor(method, p, processor);
        }
    }

    private void cleanupServerSocket() {
        if (!serverSocket.isClosed()) {
            CloseUtil.close(serverSocket);
        }
    }

    private void cleanupExecutor() {
        for (int i = 0; i < 10 && !executor.isTerminated(); ++i) {
            try {
                Logger.debug("waiting for terminating threads...");
                executor.shutdownNow();
                if (executor.awaitTermination(MAX_WAITING_TIME,
                        TimeUnit.SECONDS)) {
                    Logger.debug("done");
                    break;
                } else {
                    Logger.debug("timed out");
                }
            } catch (InterruptedException e) {
                Logger.warning(e.toString());
            }
        }
    }
}
